const NS_SVG = 'http://www.w3.org/2000/svg'

const /** @type {Record<string, string>} */ NS = {
  svg: NS_SVG,
  xhtml: 'http://www.w3.org/1999/xhtml',
  xlink: 'http://www.w3.org/1999/xlink'
}

/**
 * A helper returning constructor class of given argument.
 */
export const $classOf = (/** @type {any} */ arg) => arg?.constructor

/**
 * A helper that checks if the given argument is of type `string`.
 *
 * @returns {arg is string}
 */
export const $isString = (/** @type {any} */ arg) => $classOf(arg) === String

/**
 * A helper that checks if the given argument is of type `number`.
 *
 * @returns {arg is number}
 */
export const $isNumber = (/** @type {any} */ arg) => $classOf(arg) === Number

/**
 * A helper that checks if the given argument is of type `Record<string, any>`.
 *
 * @returns {arg is Record<string, any>}
 */
export const $isObject = (/** @type {any} */ arg) => $classOf(arg) === Object

/**
 * A helper that checks if the given argument is of type `Array`.
 */
export const $isArray = Array.isArray

/**
 * Replacement for the `in` operator (not to be confused with the `for-in` loop) that works properly.
 */
export const $in = (/** @type {any} */ key, /** @type {any} */ ref) =>
  ($isString(key) || $isNumber(key)) && Object.hasOwnProperty.call(ref ?? Object, key)

/**
 * A helper for convenient accessing to nested properties.
 */
export const $get = (/** @type {any} */ ref, /** @type {string[]} */...keys) => {
  for (const key of keys) {
    if (!$in(key, ref)) {
      return
    }
    ref = ref[key]
  }
  return ref
}

/**
 * A helper that checks the equality of the given arguments.
 *
 * @returns {boolean}
 */
export const $eq = (/** @type {any} */ x, /** @type {any} */ y) => {
  if (x === y) {
    return true
  }

  const classOfX = $classOf(x)

  if (classOfX === $classOf(y)) {
    if (classOfX === Number || classOfX === Date) {
      return (isNaN(x) && isNaN(y)) || +x === +y
    }

    if (classOfX === String || classOfX === RegExp) {
      return '' + x === '' + y
    }

    if (classOfX === Array) {
      return x.length === y.length && x.every((/** @type {any} */ item, /** @type {number} */ index) => $eq(item, y[index]))
    }

    if (classOfX === Object) {
      const keysOfX = Object.keys(x)
      return keysOfX.length === Object.keys(y).length && keysOfX.every(key => $eq(x[key], y[key]))
    }

    if (classOfX === Set || classOfX === Map) {
      if (x.size !== y.size) {
        return false
      }
      const xa = [...x]
      const ya = [...y]
      return xa.every(xv => ya.find(yv => $eq(xv, yv)))
    }
  }

  return x === y
}

/**
 * *Attributes* of the created or modified element set by `setAttribute` or `setAttributeNS`.
 *
 * @typedef {Record<string, string | number | boolean>} $Attributes
 */

/**
 * *Properties* to set on the created or modified element.
 *
 * @typedef {Record<string, any>} $Properties
 */

/**
 * *Item* to append to the created or modified element.
 *
 * @typedef {Node | string | number | $Config} $Item
 */

/**
 * *Tag* of the created element.
 *
 * @typedef {keyof HTMLElementTagNameMap | keyof SVGElementTagNameMap} $Tag
 */

/**
 * Modified *element*.
 *
 * @typedef {HTMLElement | SVGElement} $Element
 */

/**
 * @typedef {{
 *   t:  $Tag;
 *   a?: $Attributes;
 *   p?: $Properties;
 *   i?: $Item[];
 * } | {
 *   e:  $Element;
 *   a?: $Attributes;
 *   p?: $Properties;
 *   i?: $Item[];
 * }} $Config
 */

/**
 * @template {$Tag} T
 * @typedef  {{ t: T; } & $Config} $TConfig
 */

/**
 * @template {$Element} E
 * @typedef  {{ e: E; } & $Config} $EConfig
 */

/**
 * ExtJS like, lightweight helper for creating and modifying DOM elements.
 *
 * @type {{
 *   <T extends $Tag>(config: $TConfig<T>, namespace?: string):
 *       T extends keyof HTMLElementTagNameMap ? HTMLElementTagNameMap[T] :
 *       T extends keyof  SVGElementTagNameMap ?  SVGElementTagNameMap[T] :
 *       Element;
 *   <E extends $Element>(config: $EConfig<E>, namespace?: string): E;
 * }}
 */
export const $ = (
  // @ts-expect-error:
  /** @type {$TConfig<$Tag> | $EConfig<$Element>} */ { t, e, a = {}, p = {}, i = [] },
  /** @type {string | null} */ namespace = null
) => {
  if (namespace == null && t === 'svg') {
    namespace = NS_SVG
  }

  if (e == null && t != null) {
    e = namespace != null
      ? document.createElementNS(namespace, t)
      : document.createElement(t)
  }

  if (e != null) {
    for (const attribute in a) {
      const value = a[attribute]
      const indexOfColon = attribute.indexOf(':')

      if (indexOfColon >= 0) {
        const /** @type {string=} */ ns = NS[attribute.slice(0, indexOfColon)]

        if (ns != null) {
          const name = attribute.slice(indexOfColon + 1)

          if (value === true) {
            e.setAttributeNS(ns, name, '')
          } else if (value === false) {
            e.removeAttributeNS(ns, name)
          } else {
            e.setAttributeNS(ns, name, value)
          }

          continue
        }
      }

      if (value === true) {
        e.setAttribute(attribute, '')
      } else if (value === false) {
        e.removeAttribute(attribute)
      } else {
        e.setAttribute(attribute, value)
      }
    }

    for (const property in p) {
      const value = p[property]
      if (property in e && $isObject(value)) { // warning: $in cannot be used here!
        for (const key2 in value) {
          e[property][key2] = value[key2]
        }
      } else {
        e[property] = value
      }
    }

    for (const item of i) {
      const node = $isObject(item)
        // @ts-expect-error:
        ? $(item, namespace)
        : $isString(item) || $isNumber(item)
          // @ts-expect-error:
          ? document.createTextNode(item)
          : item

      e.appendChild(node)
    }
  }

  return e
}

/**
 * Syntactic sugar for `$({ t, i })`.
 *
 * @type {{
 *   <T extends $Tag>(t: T, ...i: $Item[]):
 *       T extends keyof HTMLElementTagNameMap ? HTMLElementTagNameMap[T] :
 *       T extends keyof  SVGElementTagNameMap ?  SVGElementTagNameMap[T] :
 *       Element;
 * }}
 */
export const $ti = (t, ...i) => $({ t, i })

/**
 * Syntactic sugar for `$({ t, a, i })`.
 *
 * @type {{
 *   <T extends $Tag>(t: T, a: $Attributes, ...i: $Item[]):
 *       T extends keyof HTMLElementTagNameMap ? HTMLElementTagNameMap[T] :
 *       T extends keyof  SVGElementTagNameMap ?  SVGElementTagNameMap[T] :
 *       Element;
 * }}
 */
export const $tai = (t, a, ...i) => $({ t, a, i })

/**
 * Syntactic sugar for `$({ t, p, i })`.
 *
 * @type {{
 *   <T extends $Tag>(t: T, p: $Properties, ...i: $Item[]):
 *       T extends keyof HTMLElementTagNameMap ? HTMLElementTagNameMap[T] :
 *       T extends keyof  SVGElementTagNameMap ?  SVGElementTagNameMap[T] :
 *       Element;
 * }}
 */
export const $tpi = (t, p, ...i) => $({ t, p, i })

/**
 * Syntactic sugar for `$({ t, a, p i })`.
 *
 * @type {{
 *   <T extends $Tag>(t: T, a: $Attributes, p: $Properties, ...i: $Item[]):
 *       T extends keyof HTMLElementTagNameMap ? HTMLElementTagNameMap[T] :
 *       T extends keyof  SVGElementTagNameMap ?  SVGElementTagNameMap[T] :
 *       Element;
 * }}
 */
export const $tapi = (t, a, p, ...i) => $({ t, a, p, i })

/**
 * Syntactic sugar for `$({ e, i })`.
 *
 * @type {{
 *   <E extends $Element>(e: E, ...i: $Item[]): E;
 * }}
 */
export const $ei = (e, ...i) => $({ e, i })

/**
 * Syntactic sugar for `$({ e, a, i })`.
 *
 * @type {{
 *   <E extends $Element>(e: E, a: $Attributes, ...i: $Item[]): E;
 * }}
 */
export const $eai = (e, a, ...i) => $({ e, a, i })

/**
 * Syntactic sugar for `$({ e, p, i })`.
 *
 * @type {{
 *   <E extends $Element>(e: E, p: $Properties, ...i: $Item[]): E;
 * }}
 */
export const $epi = (e, p, ...i) => $({ e, p, i })

/**
 * Syntactic sugar for `$({ e, a, p i })`.
 *
 * @type {{
 *   <E extends $Element>(e: E, a: $Attributes, p: $Properties, ...i: $Item[]): E;
 * }}
 */
export const $eapi = (e, a, p, ...i) => $({ e, a, p, i })

/**
 * A helper for creating `<a>` elements.
 */
export const $a = (/** @type {string} */ href, /** @type {$Item[]} */ ...i) => i.length > 0
  ? $tai('a', { href }, ...i)
  : $ti('code', $tai('a', { href }, href.replace(/^#/, '')))

/**
 * @typedef {{
 *   [attributeOrSelector: string]: string | number | $CSSRules;
 * }} $CSSRules
 */

/**
 * The `$CSS` format provides a hierarchical description of CSS rules.
 *
 * - Keys of sub-objects whose values are NOT objects are treated as CSS attribute, and values are treated as values of those CSS attributes; the concatenation of keys of all parent objects is a CSS rule.
 * - All keys ignore the part starting with a splitter (default: `$$`) sign until the end of the key (e.g. `src$$1` → `src`, `@font-face$$1` → `@font-face`).
 * - In keys specifying CSS attribute, all uppercase letters are replaced by lowercase letters with an additional `-` character preceding them (e.g. `fontFamily` → `font-family`).
 * - Commas in keys that makes a CSS rule cause it to “split” and create separate rules for each part (e.g. `{div:{margin:1,'.a,.b,.c':{margin:2}}}` → `div{margin:1}div.a,div.b,div.c{margin:2}`).
 * - Top-level keys that begin with `@` are not concatenated with sub-object keys.
 *
 * @typedef {{
 *   [selector: string]: $CSSRules;
 * }} $CSS
 */

const _css = (
  /** @type {$CSSRules} */ style,
  /** @type {string} */ prefix,
  /** @type {string[]} */ result,
  /** @type {(text: string) => string} */ split
) => {
  const /** @type {[$CSSRules | string[], string][]} */ queue = [[style, prefix]]

  while (queue.length > 0) {
    const [style2, prefix2] = queue.shift() ?? []

    if (style2 == null || prefix2 == null) {
      continue
    }

    if ($isArray(style2)) {
      result.push(
        prefix2,
        prefix2 !== '' ? '{' : '',
        style2.join(';'),
        prefix2 !== '' ? '}' : ''
      )
    } else {
      const /** @type {[$CSSRules | string[], string][]} */ todo = []
      let attributesPushed = false
      let /** @type {string[]} */ attributes = []

      for (const key in style2) {
        const value = style2[key]

        if ($isString(value) || $isNumber(value)) {
          if (!attributesPushed) {
            attributesPushed = true
            attributes = []
            todo.push([attributes, prefix2])
          }

          attributes.push(`${split(key).replace(/([A-Z])/g, (_, letter) => '-' + letter.toLowerCase())}:${value}`)
        } else {
          attributesPushed = false

          const /** @type {string[]} */ newPrefix = []
          const keySplitted = key.split(',')

          for (const prefixItem of prefix2.split(',')) {
            for (const keyItem of keySplitted) {
              newPrefix.push(prefixItem + keyItem)
            }
          }

          todo.push([value, newPrefix.join(',')])
        }
      }

      queue.unshift(...todo)
    }
  }
}

/**
 * CSS-in-JS helper based on the `$CSS` format.
 * An iterative implementation that does not cause a stack overflow exception.
 */
export const $css = (/** @type {$CSS} */ style, /** @type {string} */ splitter = '$$') => {
  const split = (/** @type {string} */ text) => text.split(splitter)[0]
  const /** @type {string[]} */ result = []

  for (const key in style) {
    const value = style[key]

    if (key[0] === '@') {
      result.push(split(key) + '{')
      _css(value, '', result, split)
      result.push('}')
    } else {
      _css(value, split(key), result, split)
    }
  }

  return result.join('')
}

/**
 * A generic helper for syntax highlighting.
 *
 * @returns {$Item[]}
 */
export const $highlight = (
  /** @type {RegExp} */ pattern,
  /** @type {((chunk: string, index: number) => $Item)[]} */ highlighters,
  /** @type {string} */ innerText
) => {
  const /** @type {$Item[]} */ result = []

  innerText.split(pattern).forEach((chunk, index) => {
    index %= highlighters.length
    if (chunk != null) {
      result.push(highlighters[index](chunk, index))
    }
  })

  return result
}

/**
 * A helper for highlighting JavaScript syntax.
 */
// @ts-expect-error:
export const $highlightJs = $highlight.bind(0,
  /('.*?'|".*?"|`[\s\S]*?`)|(\/\/.*?\n|\/\*[\s\S]*?\*\/)|(break|case|catch|const|continue|debugger|default|delete|do|else|eval|export\s+type|export|extends|false|finally|for|from|function|goto|if|import|in|instanceof|is|keyof|let|NaN|new|null|package|return|super|switch|this|throw|true|try|typeof|undefined|var|void|while|with|yield)(?!\w)|([<>=.?:&|!~*/%+-])|(\d+(?:\.[\d_]+)?(?:e[+-]?[\d_]+)?)|([$\w]+)(?=\()|([$\wąćęłńóśżźĄĆĘŁŃÓŚŻŹ]+)/g,
  [
    chunk => chunk, // no match
    chunk => $ti('u', chunk), // string literals
    chunk => $ti('em', chunk), // comments
    chunk => $ti('b', chunk), // keywords
    chunk => $ti('b', chunk), // operators
    chunk => $ti('u', chunk), // number literals
    chunk => $ti('u', chunk), // function calls
    chunk => $ti('i', chunk) // literals
  ]
)

/** @returns {string} */
const _locale = (
  /** @type {Record<string, Record<string, string | Record<string, string>>>} */ locales,
  /** @type {string} */ language,
  /** @type {string} */ text,
  /** @type {string} */ version
) => $get(locales, language, version, text) ?? $get(locales, language, text) ?? text

/**
 * Language translations helper.
 */
export const $locale = (
  /** @type {Record<string, Record<string, string | Record<string, string>>>} */ locales,
  /** @type {string} */ defaultLanguage,
  /** @type {readonly string[]} */ languages = navigator.languages
) => {
  for (const language of languages) {
    if ($in(language, locales)) {
      // @ts-expect-error:
      return _locale.bind(0, locales, language)
    }
  }

  // @ts-expect-error:
  return _locale.bind(0, locales, defaultLanguage)
}
